/************************************************************
* Copyright (C) 2006-2008 Masahiko SAWAI All Rights Reserved. 
************************************************************/

#include "wiiremote.h"
#include "wiiremote_impl.h"
#include "DebugLog.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#define DEFAULT_SPEAKER_FORMAT 0x00 
#define DEFAULT_SPEAKER_VOLUME 0x40 /* 0 - 0xff */
#define DEFAULT_SPEAKER_SAMPLE_RATE WRMT_SPEAKER_SAMPLE_RATE_3920

#define OUTPUT_RETRY_COUNT 3
#define OUTPUT_RETRY_WAIT 10 /* ms */

#define ADDRESS_SPEAKER_CONFIG

/* type definitions */
struct wrmt_wiiremote
{
	WRMT_WiiRemoteImpl *impl;
	int device_index;
	/* buffers */
	unsigned char *output_buffer;
	unsigned char *input_buffer;
	int functions[WRMT_NUMBER_OF_FUNCTIONS];
	int device_state[WRMT_NUMBER_OF_DATA_TYPE];
	unsigned char function_updated_flags[WRMT_NUMBER_OF_FUNCTIONS];
	unsigned char device_state_updated_flags[WRMT_NUMBER_OF_DATA_TYPE];
};

#define WRMT_WiiRemote_Invaliant(self) \
{ \
	assert((self) != NULL); \
	assert((self)->impl != NULL); \
	assert(((self)->device_index >= 0) && \
		((self)->device_index < WRMT_MAX_DEVICES)); \
	assert((self)->output_buffer != NULL); \
	assert((self)->input_buffer != NULL); \
	assert( ((self)->functions[WRMT_FUNCTION_CONTINUOUS] == 0) || \
		((self)->functions[WRMT_FUNCTION_CONTINUOUS] == 1)); \
	assert( ((self)->functions[WRMT_FUNCTION_MOTION] == 0) || \
		((self)->functions[WRMT_FUNCTION_MOTION] == 1)); \
	assert( ((self)->functions[WRMT_FUNCTION_IR] == 0) || \
		((self)->functions[WRMT_FUNCTION_IR] == 1)); \
	assert( ((self)->functions[WRMT_FUNCTION_SPEAKER] == 0) || \
		((self)->functions[WRMT_FUNCTION_SPEAKER] == 1)); \
	assert( ((self)->device_state[WRMT_DATA_FORCE_FEEDBACK] == 0) || \
		((self)->device_state[WRMT_DATA_FORCE_FEEDBACK] == 1)); \
	assert( ((self)->device_state[WRMT_DATA_LEDS] >= 0) || \
		((self)->device_state[WRMT_DATA_LEDS] <= 0x0f)); \
} 

#define WRMT_WiiRemote_ValidateOutputReport(self) \
{ \
	assert( ((self)->output_buffer[0] >= WRMT_OUTPUT_REPORT_ID_FIRST) && \
		((self)->output_buffer[0] <= WRMT_OUTPUT_REPORT_ID_LAST)); \
} 


/************************************************************
#private variables
************************************************************/
static
const
unsigned char
IR_SENSITIVITY_MIDRANGE_DATA_1[] = 
	{0x02, 0x00, 0x00, 0x71, 0x01, 0x00, 0xaa, 0x00, 0x64};

static
const
unsigned char
IR_SENSITIVITY_MIDRANGE_DATA_2[] =
	{0x63, 0x03};

static
WRMT_WiiRemote
wiiRemoteList[WRMT_MAX_DEVICES];

static
const 
char *errorMessage = NULL;


/************************************************************
#private functions
************************************************************/

static
void
WRMT_WiiRemote_InitWithDeviceIndex(WRMT_WiiRemote *self,
	int device_index)
{
	int i;
	WRMT_WiiRemoteImpl *wiiRemoteImpl;
	assert(self != NULL);
	assert(device_index >= 0 && device_index <= WRMT_MAX_DEVICES);
	DebugLog("Hello\n");

	wiiRemoteImpl = WRMT_Impl_GetWiiRemoteAt(device_index);
	self->impl = wiiRemoteImpl;
	self->device_index = device_index;
	self->output_buffer = WRMT_WiiRemoteImpl_GetOutputBuffer(wiiRemoteImpl);
	self->input_buffer = WRMT_WiiRemoteImpl_GetInputBuffer(wiiRemoteImpl);

	for (i = 0;i < WRMT_NUMBER_OF_FUNCTIONS;i++)
	{
		self->functions[i] = 0;
	}

	for (i = 0;i < WRMT_NUMBER_OF_DATA_TYPE;i++)
	{
		self->device_state[i] = 0;
	}
	self->device_state[WRMT_DATA_SPEAKER_FORMAT] = DEFAULT_SPEAKER_FORMAT;
	self->device_state[WRMT_DATA_SPEAKER_VOLUME] = DEFAULT_SPEAKER_VOLUME;
	self->device_state[WRMT_DATA_SPEAKER_SAMPLE_RATE] = DEFAULT_SPEAKER_SAMPLE_RATE;

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

static
void
WRMT_WiiRemote_ParseBatteryLevel(WRMT_WiiRemote *self)
{
	assert(self->input_buffer[0] == 0x20);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	self->device_state[WRMT_DATA_BATTERY_LEVEL] = self->input_buffer[6];

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

static
void
WRMT_WiiRemote_ParseButtonState(WRMT_WiiRemote *self)
{
	int button_state;
	assert(self->input_buffer[0] >= 0x30 && self->input_buffer[0] <= 0x3f);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	button_state = (self->input_buffer[1] << 8) & 0xff00;
	button_state |= self->input_buffer[2] & 0x00ff;
	self->device_state[WRMT_DATA_BUTTONS] = button_state;

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

static
void
WRMT_WiiRemote_ParseMotionState(WRMT_WiiRemote *self)
{
	assert(self->input_buffer[0] & 0x01);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	self->device_state[WRMT_DATA_MOTION_X] = self->input_buffer[3] & 0xff;
	self->device_state[WRMT_DATA_MOTION_Y] = self->input_buffer[4] & 0xff;
	self->device_state[WRMT_DATA_MOTION_Z] = self->input_buffer[5] & 0xff;

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

static
void
WRMT_WiiRemote_ParseIRState(WRMT_WiiRemote *self)
{
	unsigned char *input_buffer = self->input_buffer;
	int *device_state = self->device_state;
	assert(self->input_buffer[0] & 0x02);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	if (input_buffer[6] == 0xff &&
		input_buffer[7] == 0xff &&
		input_buffer[8] == 0xff)
	{
		DebugLog("IR1 not found\n");
		device_state[WRMT_DATA_IR1_FOUND] = 0;
		device_state[WRMT_DATA_IR1_SIZE] = 0;
		device_state[WRMT_DATA_IR1_X] = 0;
		device_state[WRMT_DATA_IR1_Y] = 0;
	}
	else
	{
		DebugLog("IR1 found\n");
		device_state[WRMT_DATA_IR1_FOUND] = 1;
		device_state[WRMT_DATA_IR1_SIZE] = input_buffer[ 8] & 0x0f;
		device_state[WRMT_DATA_IR1_X] = input_buffer[ 6];
		device_state[WRMT_DATA_IR1_Y] = input_buffer[ 7];
		device_state[WRMT_DATA_IR1_X] |= ((int)input_buffer[ 8] & 0x30) << 4;
		device_state[WRMT_DATA_IR1_Y] |= ((int)input_buffer[ 8] & 0xc0) << 2;
	}

	if (input_buffer[9] == 0xff &&
		input_buffer[10] == 0xff &&
		input_buffer[11] == 0xff)
	{
		DebugLog("IR2 not found\n");
		device_state[WRMT_DATA_IR2_FOUND] = 0;
		device_state[WRMT_DATA_IR2_SIZE] = 0;
		device_state[WRMT_DATA_IR2_X] = 0;
		device_state[WRMT_DATA_IR2_Y] = 0;
	}
	else
	{
		DebugLog("IR2 found\n");
		device_state[WRMT_DATA_IR2_FOUND] = 1;
		device_state[WRMT_DATA_IR2_SIZE] = input_buffer[11] & 0x0f;
		device_state[WRMT_DATA_IR2_X] = input_buffer[ 9];
		device_state[WRMT_DATA_IR2_Y] = input_buffer[10];
		device_state[WRMT_DATA_IR2_X] |= ((int)input_buffer[11] & 0x30) << 4;
		device_state[WRMT_DATA_IR2_Y] |= ((int)input_buffer[11] & 0xc0) << 2;
	}

	if (device_state[WRMT_DATA_IR1_FOUND] &&
		device_state[WRMT_DATA_IR2_FOUND])
	{
		int size, x, y;

		size = device_state[WRMT_DATA_IR1_SIZE] + device_state[WRMT_DATA_IR2_SIZE];
		x = device_state[WRMT_DATA_IR1_X] + device_state[WRMT_DATA_IR2_X];
		y = device_state[WRMT_DATA_IR1_Y] + device_state[WRMT_DATA_IR2_Y];
		device_state[WRMT_DATA_IR_FOUND] = 1;
		device_state[WRMT_DATA_IR_SIZE] = size / 2;
		device_state[WRMT_DATA_IR_X] = x / 2;
		device_state[WRMT_DATA_IR_Y] = y / 2;
	}
	else
	{
		device_state[WRMT_DATA_IR_FOUND] = 0;
		device_state[WRMT_DATA_IR_SIZE] = 0;
		device_state[WRMT_DATA_IR_X] = 0;
		device_state[WRMT_DATA_IR_Y] = 0;
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

static
void
WRMT_WiiRemote_ParseReport(WRMT_WiiRemote *self)
{
	unsigned char report_id;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

#ifdef DUMP_INPUT_BUFFER
	{
		int i;
		fprintf(stderr, "input_buffer : ");
		for (i = 0;i < WRMT_BUFFER_SIZE;i++)
			fprintf(stderr, "%02x ", self->input_buffer[i]);
		fprintf(stderr, "\n");
	}
#endif

	report_id = self->input_buffer[0];
	DebugLog("report_id : 0x%02x\n", report_id);
	if ((report_id & 0xf0) == 0x30)
	{
		WRMT_WiiRemote_ParseButtonState(self);
		if (report_id & 0x01)
		{
			WRMT_WiiRemote_ParseMotionState(self);
		}

		if ((report_id & 0x02) &&
			self->functions[WRMT_FUNCTION_IR])
		{
			WRMT_WiiRemote_ParseIRState(self);
		}
	}
	else if (report_id == 0x20)
	{ /* Expansion port and controler battery status */
		WRMT_WiiRemote_ParseBatteryLevel(self);
	}

	/* clear report ID */
	self->input_buffer[0] = 0;

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

static
void
WRMT_WiiRemote_ClearOutputBuffer(WRMT_WiiRemote *self)
{
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	memset(self->output_buffer, 0, WRMT_BUFFER_SIZE);

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

static
WRMT_IOReturn
WRMT_WiiRemote_OutputToDevice(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result = WRMT_IO_SUCCESS;
	int retry_count = OUTPUT_RETRY_COUNT;
	WRMT_WiiRemote_Invaliant(self);
	WRMT_WiiRemote_ValidateOutputReport(self);
	DebugLog("Hello\n");

#ifdef DUMP_OUTPUT_BUFFER
	{
		int i;
		fprintf(stderr, "output_buffer : ");
		for (i = 0;i < WRMT_BUFFER_SIZE;i++)
			fprintf(stderr, "%02x ", self->output_buffer[i]);
		fprintf(stderr, "\n");
	}
#endif

	for (retry_count = OUTPUT_RETRY_COUNT;retry_count > 0;retry_count--)
	{
		result = WRMT_WiiRemoteImpl_OutputToDevice(self->impl);
		DebugLog("retry_count : %d, result : %d\n", retry_count, result);
		if (result == WRMT_IO_SUCCESS) break;
		WRMT_Sleep(OUTPUT_RETRY_WAIT);
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}

static
WRMT_IOReturn
WRMT_WiiRemote_SendControllerStatusRequest(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_CONTROLLER_STATUS;
	self->output_buffer[1] = 0x00;
	result = WRMT_WiiRemote_OutputToDevice(self);

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}

/**
* update report type. This function refer following attributes.
* - self->functions[WRMT_FUNCTION_CONTINUOUS]
* - self->functions[WRMT_FUNCTION_MOTION]
* - self->functions[WRMT_FUNCTION_IR]
* - self->device_state[WRMT_DATA_FORCE_FEEDBACK]
**/
static
WRMT_IOReturn
WRMT_WiiRemote_SendReportTypeRequest(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result;
	int channel, continuous, force_feedback;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	continuous = 0x00;
	if (self->functions[WRMT_FUNCTION_CONTINUOUS] ||
		self->functions[WRMT_FUNCTION_IR])
	{
		continuous = 0x04;
	}
	force_feedback = self->device_state[WRMT_DATA_FORCE_FEEDBACK];

	channel = 0x30;
	if (self->functions[WRMT_FUNCTION_MOTION])
		channel |= 0x01;
	if (self->functions[WRMT_FUNCTION_IR])
		channel |= 0x03;

	DebugLog("channel : 0x%02x\n", channel);
	WRMT_WiiRemote_ClearOutputBuffer(self);
	self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_REQUEST_REPORT_TYPE;
	self->output_buffer[1] = continuous | force_feedback;
	self->output_buffer[2] = channel;
	result = WRMT_WiiRemote_OutputToDevice(self);

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}


/**
* Send force feedback state and LED state. This function refer following attributes.
* - self->device_state[WRMT_DATA_FORCE_FEEDBACK]
* - self->device_state[WRMT_DATA_LEDS]
**/
static
WRMT_IOReturn
WRMT_WiiRemote_SendForceFeedbackAndLED(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result;
	WRMT_WiiRemote_Invaliant(self);

	WRMT_WiiRemote_ClearOutputBuffer(self);
	self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_LED_AND_FORCE_FEEDBACK;
	self->output_buffer[1] = 0;
	self->output_buffer[1] |= self->device_state[WRMT_DATA_FORCE_FEEDBACK];
	self->output_buffer[1] |= (self->device_state[WRMT_DATA_LEDS] << 4);
	result = WRMT_WiiRemote_OutputToDevice(self);

	WRMT_WiiRemote_Invaliant(self);
	return result;
}

/**
* Send IR Parameters. This function refer following attributes.
* - self->functions[WRMT_FUNCTION_IR]
* - self->device_state[WRMT_DATA_FORCE_FEEDBACK]
**/
static
WRMT_IOReturn
WRMT_WiiRemote_SendIRParameters(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result;
	int ir_flag, force_feedback_flag;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	ir_flag = (self->functions[WRMT_FUNCTION_IR] == 0) ? 0x00 : 0x04;
	force_feedback_flag = self->device_state[WRMT_DATA_FORCE_FEEDBACK];

	/* IR Sensor Enable */
	WRMT_WiiRemote_ClearOutputBuffer(self);
	self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_IR_SENSOR_ENABLE;
	self->output_buffer[1] = ir_flag | force_feedback_flag;
	result = WRMT_WiiRemote_OutputToDevice(self);
	if (result != WRMT_IO_SUCCESS) return result;

	/* IR Sensor Enable 2 */
	WRMT_WiiRemote_ClearOutputBuffer(self);
	self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_IR_SENSOR_ENABLE2;
	self->output_buffer[1] = ir_flag | force_feedback_flag;
	result = WRMT_WiiRemote_OutputToDevice(self);
	if (result != WRMT_IO_SUCCESS) return result;

	if (self->functions[WRMT_FUNCTION_IR])
	{
		unsigned char disable_ir = 0x01, enable_ir = 0x08;
		unsigned char ir_mode = 0x33;

		/* stop ir camera */
		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04b00030, &disable_ir, 1);
		if (result != WRMT_IO_SUCCESS) return result;

		/* set ir sensitivity */
		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04b00000, IR_SENSITIVITY_MIDRANGE_DATA_1, 9);
		if (result != WRMT_IO_SUCCESS) return result;

		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04b0001a, IR_SENSITIVITY_MIDRANGE_DATA_2, 2);
		if (result != WRMT_IO_SUCCESS) return result;

		/* set ir mode */
		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04b00033, &ir_mode, 1);
		if (result != WRMT_IO_SUCCESS) return result;

		/* start ir camera */
		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04b00030, &enable_ir, 1);
		if (result != WRMT_IO_SUCCESS) return result;
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}

/**
* Send Speaker Parameters. This function refer following attributes.
* - self->device_state[WRMT_DATA_SPEAKER_FORMAT]
* - self->device_state[WRMT_DATA_SPEAKER_SAMPLE_RATE]
* - self->device_state[WRMT_DATA_SPEAKER_VOLUME]
**/
static
WRMT_IOReturn
WRMT_WiiRemote_SendSpeakerParameters(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result;
	unsigned char speaker_config[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	/* sound format  */
	speaker_config[1] = self->device_state[WRMT_DATA_SPEAKER_FORMAT] & 0xff;

	/* sample rate */
	speaker_config[3] = self->device_state[WRMT_DATA_SPEAKER_SAMPLE_RATE] & 0xff;

	/* volume (0-255) */
	speaker_config[4] = self->device_state[WRMT_DATA_SPEAKER_VOLUME] & 0xff;

	/* write speaker_config */
	result = WRMT_WiiRemote_WriteDataToMemory(self, 
		0x04a20001 , speaker_config, 7);

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}

/**
* Send Speaker Parameters. This function refer following attributes.
* - self->functions[WRMT_FUNCTION_SPEAKER]
* - self->device_state[WRMT_DATA_SPEAKER_FORMAT]
* - self->device_state[WRMT_DATA_SPEAKER_SAMPLE_RATE]
* - self->device_state[WRMT_DATA_SPEAKER_VOLUME]
**/
static
WRMT_IOReturn
WRMT_WiiRemote_SetUpSpeaker(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	/* enable/disable speaker */
	WRMT_WiiRemote_ClearOutputBuffer(self);
	self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_SPEAKER_ENABLE;
	self->output_buffer[1] = self->functions[WRMT_FUNCTION_SPEAKER] ? 0x04 : 0x00;
	result = WRMT_WiiRemote_OutputToDevice(self);
	if (result != WRMT_IO_SUCCESS) return result;

	if (self->functions[WRMT_FUNCTION_SPEAKER])
	{
		const unsigned char data1 = 0x01, data8 = 0x08;

		/* mute speaker */
		WRMT_WiiRemote_ClearOutputBuffer(self);
		self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_SPEAKER_MUTE;
		self->output_buffer[1] = 0x04;
		result = WRMT_WiiRemote_OutputToDevice(self);
		if (result != WRMT_IO_SUCCESS) return result;

		/* setup speaker */
		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04a20009, &data1, 1);
		if (result != WRMT_IO_SUCCESS) return result;
		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04a20001, &data8, 1);
		if (result != WRMT_IO_SUCCESS) return result;

		/* send parameters */
		WRMT_WiiRemote_SendSpeakerParameters(self);

		/* setup speaker */
		result = WRMT_WiiRemote_WriteDataToMemory(self, 
			0x04a20008, &data1, 1);
		if (result != WRMT_IO_SUCCESS) return result;

		/* unmute speaker */
		WRMT_WiiRemote_ClearOutputBuffer(self);
		self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_SPEAKER_MUTE;
		self->output_buffer[1] = 0x00;
		result = WRMT_WiiRemote_OutputToDevice(self);
		if (result != WRMT_IO_SUCCESS) return result;
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}


/************************************************************
#public functions
************************************************************/

/**
* FIXME
**/
void WRMT_SetError(const char *message)
{
	errorMessage = message;
}

/**
* FIXME
**/
char *
WRMT_GetError()
{
	return (char *)errorMessage;
}

int
WRMT_Init()
{
	int result;
	DebugLog("Hello\n");

	result = WRMT_Impl_Init(3);
	if (result == 0)
	{
		int number, device_index;

		number = WRMT_Impl_GetNumWiiRemote();
		for (device_index = 0;device_index < number;device_index++)
		{
			WRMT_WiiRemote *wiiRemote = &(wiiRemoteList[device_index]);
			WRMT_WiiRemote_InitWithDeviceIndex(wiiRemote, device_index);
		}
	}

	DebugLog("Bye\n");
	assert(result == 0 || result == -1);
	return result;
}

void
WRMT_Quit()
{
	DebugLog("Hello\n");

	WRMT_Impl_Quit();

	DebugLog("Bye\n");
}

void
WRMT_Sleep(int ms)
{
	DebugLog("Hello\n");

	WRMT_Impl_Sleep(ms);

	DebugLog("Bye\n");
}

int
WRMT_UpdateAll()
{
	WRMT_IOReturn result;
	WRMT_IOReturn rc;
	int i;
	DebugLog("Hello\n");

	/* send updated state to device*/
	for (i = WRMT_GetNumWiiRemote()-1;i >= 0 ;i--)
	{
		WRMT_WiiRemote *wiiRemote = &(wiiRemoteList[i]);
		if (WRMT_WiiRemote_IsOpened(wiiRemote))
		{
			WRMT_WiiRemote_Update(wiiRemote);
		}
	}

	/* receive all event from device */
	do
	{
		rc = WRMT_Poll();
	}
	while (rc >= 0);

	result = (rc == WRMT_IO_TIMEOUT) ? WRMT_IO_SUCCESS : WRMT_IO_ERROR;
	
	DebugLog("Bye\n");
	assert(result == WRMT_IO_SUCCESS || result == WRMT_IO_ERROR);
	return result;
}

int
WRMT_Poll()
{
	int result;
	WRMT_IOReturn rc;
	WRMT_WiiRemote *wiiRemote;
	int updated_device_index;
	DebugLog("Hello\n");

	/* Polling devices*/
	rc = WRMT_Impl_Poll(&updated_device_index);

	/* Update device status*/
	if(rc == WRMT_IO_SUCCESS)
	{
		wiiRemote = &(wiiRemoteList[updated_device_index]);
		WRMT_WiiRemote_ParseReport(wiiRemote);

		result = updated_device_index;
	}
	else
	{
		result = rc;
	}

	DebugLog("Bye\n");
	assert(result == WRMT_IO_ERROR || result == WRMT_IO_TIMEOUT || result >= 0);
	return result;
}

int
WRMT_GetNumWiiRemote()
{
	return WRMT_Impl_GetNumWiiRemote();
}

WRMT_WiiRemote *
WRMT_GetWiiRemoteAt(int device_index)
{
	WRMT_WiiRemote *result = NULL;
	assert(device_index >= 0 &&
		device_index < WRMT_Impl_GetNumWiiRemote());
	DebugLog("Hello device_index = %d\n", device_index);

	result = &(wiiRemoteList[device_index]);

	DebugLog("Bye\n");
	assert(result != NULL);
	return result;
}

WRMT_IOReturn
WRMT_WiiRemote_Open(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello self = %p\n", self);

	result = WRMT_WiiRemoteImpl_Open(self->impl);
	if (result == WRMT_IO_SUCCESS)
	{
		WRMT_WiiRemote_SendReportTypeRequest(self);
		WRMT_WiiRemote_SendForceFeedbackAndLED(self);
		WRMT_WiiRemote_SendControllerStatusRequest(self);
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}

int
WRMT_WiiRemote_IsOpened(WRMT_WiiRemote *self)
{
	int result = 0;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	result = WRMT_WiiRemoteImpl_IsOpened(self->impl);

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	return result;
}

void
WRMT_WiiRemote_Close(WRMT_WiiRemote *self)
{
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	WRMT_WiiRemoteImpl_Close(self->impl);

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}


int 
WRMT_WiiRemote_IsEnabled(WRMT_WiiRemote *self, WRMT_FunctionType function_type)
{
	int result;
	assert(function_type >= 0 && function_type < WRMT_NUMBER_OF_FUNCTIONS);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello function_type == %d\n", function_type);

	result = self->functions[function_type];

	DebugLog("Bye result == %d\n", result);
	WRMT_WiiRemote_Invaliant(self);
	return result;
}

void
WRMT_WiiRemote_SetEnabled(WRMT_WiiRemote *self,
	WRMT_FunctionType function_type, int value)
{
	assert(function_type >= 0 && function_type < WRMT_NUMBER_OF_FUNCTIONS);
	assert(value == 0 || value == 1);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello: function_type == %d, value == %d\n", function_type, value);

	if (self->functions[function_type] != value)
	{
		self->functions[function_type] = value;
		self->function_updated_flags[function_type] = 1;
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

int
WRMT_WiiRemote_GetState(WRMT_WiiRemote *self, WRMT_DataType data_type)
{
	assert(data_type >= 0 && data_type < WRMT_NUMBER_OF_DATA_TYPE);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	return self->device_state[data_type];
}

void
WRMT_WiiRemote_SetState(WRMT_WiiRemote *self, WRMT_DataType data_type, int value)
{
	assert(data_type == WRMT_DATA_FORCE_FEEDBACK ||
		data_type == WRMT_DATA_LEDS ||
		data_type == WRMT_DATA_SPEAKER_FORMAT ||
		data_type == WRMT_DATA_SPEAKER_VOLUME ||
		data_type == WRMT_DATA_SPEAKER_SAMPLE_RATE);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello data_type == %d, value == %d\n", data_type, value);

	if (self->device_state[data_type] != value)
	{
		self->device_state[data_type] = value;
		self->device_state_updated_flags[data_type] = 1;
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
}

WRMT_IOReturn
WRMT_WiiRemote_Update(WRMT_WiiRemote *self)
{
	WRMT_IOReturn result = WRMT_IO_SUCCESS;
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	/* update functions */
	if ( self->function_updated_flags[WRMT_FUNCTION_CONTINUOUS] ||
		self->function_updated_flags[WRMT_FUNCTION_MOTION] ||
		self->function_updated_flags[WRMT_FUNCTION_IR])
	{
		result = WRMT_WiiRemote_SendReportTypeRequest(self);
		if (result != WRMT_IO_SUCCESS) return result;

		if (self->function_updated_flags[WRMT_FUNCTION_IR])
		{
			result = WRMT_WiiRemote_SendIRParameters(self);
			if (result != WRMT_IO_SUCCESS) return result;
		}

		self->function_updated_flags[WRMT_FUNCTION_CONTINUOUS] = 0;
		self->function_updated_flags[WRMT_FUNCTION_MOTION] = 0;
		self->function_updated_flags[WRMT_FUNCTION_IR] = 0;
	}

	if (self->function_updated_flags[WRMT_FUNCTION_SPEAKER])
	{
		result = WRMT_WiiRemote_SetUpSpeaker(self);
		if (result != WRMT_IO_SUCCESS) return result;

		self->function_updated_flags[WRMT_FUNCTION_SPEAKER] = 0;
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_FORMAT] = 0;
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_VOLUME] = 0;
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_SAMPLE_RATE] = 0;
	}

	/* update device_state */
	if (self->device_state_updated_flags[WRMT_DATA_FORCE_FEEDBACK] ||
		self->device_state_updated_flags[WRMT_DATA_LEDS])
	{
		result = WRMT_WiiRemote_SendForceFeedbackAndLED(self);
		if (result != WRMT_IO_SUCCESS) return result;

		self->device_state_updated_flags[WRMT_DATA_FORCE_FEEDBACK] = 0;
		self->device_state_updated_flags[WRMT_DATA_LEDS] = 0;
	}

	if (self->device_state_updated_flags[WRMT_DATA_SPEAKER_FORMAT] ||
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_VOLUME] ||
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_SAMPLE_RATE])
	{
		result = WRMT_WiiRemote_SendSpeakerParameters(self);
		if (result != WRMT_IO_SUCCESS) return result;
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_FORMAT] = 0;
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_VOLUME] = 0;
		self->device_state_updated_flags[WRMT_DATA_SPEAKER_SAMPLE_RATE] = 0;
	}
	
	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	assert(result == WRMT_IO_ERROR || result == WRMT_IO_SUCCESS);
	return result;
}

WRMT_IOReturn
WRMT_WiiRemote_WriteDataToMemory(WRMT_WiiRemote *self,
	int address, const unsigned char* data_ptr, size_t data_size)
{
	WRMT_IOReturn result;
	WRMT_WiiRemote_Invaliant(self);
	assert(data_ptr != NULL);
	assert(data_size >=1 && data_size <= 16);
	DebugLog("Hello\n");

	if (data_size >=1 && data_size <= 16)
	{
		WRMT_WiiRemote_ClearOutputBuffer(self);
		self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_WRITE_DATA;
		self->output_buffer[1] = (address >> 24 & 0xFF);
		self->output_buffer[2] = (address >> 16 & 0xFF);
		self->output_buffer[3] = (address >> 8  & 0xFF);
		self->output_buffer[4] = (address & 0xFF);
		self->output_buffer[5] = (unsigned char)data_size;
		memcpy(self->output_buffer + 6, data_ptr, data_size);
		result = WRMT_WiiRemote_OutputToDevice(self);
	}
	else
	{
		WRMT_SetError("WRMT_WiiRemote_WriteDataToMemory() Illegal data_size.");
		result = WRMT_IO_ERROR;
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	assert(result == WRMT_IO_ERROR || result == WRMT_IO_SUCCESS);
	return result;
}

WRMT_IOReturn
WRMT_WiiRemote_PlaySound(WRMT_WiiRemote *self,
	const unsigned char *data, size_t data_size)
{
	WRMT_IOReturn result = WRMT_IO_SUCCESS;
	const unsigned char *data_pointer = data;
	size_t written_size = 0;
	assert(data != NULL);
	assert(data_size > 0);
	WRMT_WiiRemote_Invaliant(self);
	DebugLog("Hello\n");

	while (written_size < data_size)
	{
		size_t size;

		size = data_size - written_size;
		if (size > 20) size = 20;

		WRMT_WiiRemote_ClearOutputBuffer(self);
		self->output_buffer[0] = WRMT_OUTPUT_REPORT_ID_SPEAKER_DATA;
		self->output_buffer[1] = (unsigned char)(size << 3);
		memcpy(&(self->output_buffer[2]), data_pointer, size);

		result = WRMT_WiiRemote_OutputToDevice(self);
		if (result != WRMT_IO_SUCCESS) break;

		written_size += size;
		data_pointer += size;
	}

	DebugLog("Bye\n");
	WRMT_WiiRemote_Invaliant(self);
	assert(result == WRMT_IO_ERROR || result == WRMT_IO_SUCCESS);
	return result;
}
